@@ -1,11 +1,19 @@
 # frozen_string_literal: true
 
-require 'minitest/global_expectations/autorun'
-require 'rack'
-require 'rack/response'
-require 'stringio'
+require_relative 'helper'
 
 describe Rack::Response do
+  it 'has standard constructor' do
+    headers = { "header" => "value" }
+    body = ["body"]
+
+    response = Rack::Response[200, headers, body]
+
+    response.status.must_equal 200
+    response.headers.must_equal headers
+    response.body.must_equal body
+  end
+
   it 'has cache-control methods' do
     response = Rack::Response.new
     cc = 'foo'
@@ -22,6 +30,14 @@ describe Rack::Response do
     assert_equal etag, response.to_a[1]['ETag']
   end
 
+  it 'has a content-type method' do
+    response = Rack::Response.new
+    content_type = 'foo'
+    response.content_type = content_type
+    assert_equal content_type, response.content_type
+    assert_equal content_type, response.to_a[1]['Content-Type']
+  end
+
   it "have sensible default values" do
     response = Rack::Response.new
     status, header, body = response.finish
@@ -40,12 +56,11 @@ describe Rack::Response do
     }
   end
 
-  it "can be written to" do
-    response = Rack::Response.new
+  it "can be written to inside finish block, but does not update Content-Length" do
+    response = Rack::Response.new('foo')
+    response.write "bar"
 
-    _, _, body = response.finish do
-      response.write "foo"
-      response.write "bar"
+    _, h, body = response.finish do
       response.write "baz"
     end
 
@@ -53,6 +68,7 @@ describe Rack::Response do
     body.each { |part| parts << part }
 
     parts.must_equal ["foo", "bar", "baz"]
+    h['Content-Length'].must_equal '6'
   end
 
   it "can set and read headers" do
@@ -63,13 +79,13 @@ describe Rack::Response do
   end
 
   it "doesn't mutate given headers" do
-    [{}, Rack::Utils::HeaderHash.new].each do |header|
-      response = Rack::Response.new([], 200, header)
-      response.header["Content-Type"] = "text/plain"
-      response.header["Content-Type"].must_equal "text/plain"
+    headers = {}
 
-      header.wont_include("Content-Type")
-    end
+    response = Rack::Response.new([], 200, headers)
+    response.headers["Content-Type"] = "text/plain"
+    response.headers["Content-Type"].must_equal "text/plain"
+
+    headers.wont_include("Content-Type")
   end
 
   it "can override the initial Content-Type with a different case" do
@@ -232,6 +248,18 @@ describe Rack::Response do
                                          "foo=; domain=sample.example.com; max-age=0; expires=Thu, 01 Jan 1970 00:00:00 GMT"].join("\n")
   end
 
+  it "only deletes cookies for the domain specified" do
+    response = Rack::Response.new
+    response.set_cookie "foo", { value: "bar", domain: "example.com.example.com" }
+    response.set_cookie "foo", { value: "bar", domain: "example.com" }
+    response["Set-Cookie"].must_equal ["foo=bar; domain=example.com.example.com", "foo=bar; domain=example.com"].join("\n")
+    response.delete_cookie "foo", domain: "example.com"
+    response["Set-Cookie"].must_equal ["foo=bar; domain=example.com.example.com", "foo=; domain=example.com; max-age=0; expires=Thu, 01 Jan 1970 00:00:00 GMT"].join("\n")
+    response.delete_cookie "foo", domain: "example.com.example.com"
+    response["Set-Cookie"].must_equal ["foo=; domain=example.com; max-age=0; expires=Thu, 01 Jan 1970 00:00:00 GMT",
+                                         "foo=; domain=example.com.example.com; max-age=0; expires=Thu, 01 Jan 1970 00:00:00 GMT"].join("\n")
+  end
+
   it "can delete cookies with the same name with different paths" do
     response = Rack::Response.new
     response.set_cookie "foo", { value: "bar", path: "/" }
@@ -244,6 +272,71 @@ describe Rack::Response do
                                          "foo=; path=/path; max-age=0; expires=Thu, 01 Jan 1970 00:00:00 GMT"].join("\n")
   end
 
+  it "only delete cookies with the path specified" do
+    response = Rack::Response.new
+    response.set_cookie "foo", value: "bar", path: "/"
+    response.set_cookie "foo", value: "bar", path: "/a"
+    response.set_cookie "foo", value: "bar", path: "/a/b"
+    response["Set-Cookie"].must_equal ["foo=bar; path=/",
+                                       "foo=bar; path=/a",
+                                       "foo=bar; path=/a/b"].join("\n")
+
+    response.delete_cookie "foo", path: "/a"
+    response["Set-Cookie"].must_equal ["foo=bar; path=/",
+                                       "foo=bar; path=/a/b",
+                                       "foo=; path=/a; max-age=0; expires=Thu, 01 Jan 1970 00:00:00 GMT"].join("\n")
+  end
+
+  it "only delete cookies with the domain and path specified" do
+    response = Rack::Response.new
+    response.set_cookie "foo", value: "bar", path: "/"
+    response.set_cookie "foo", value: "bar", path: "/a"
+    response.set_cookie "foo", value: "bar", path: "/a/b"
+    response.set_cookie "foo", value: "bar", path: "/", domain: "example.com.example.com"
+    response.set_cookie "foo", value: "bar", path: "/a", domain: "example.com.example.com"
+    response.set_cookie "foo", value: "bar", path: "/a/b", domain: "example.com.example.com"
+    response.set_cookie "foo", value: "bar", path: "/", domain: "example.com"
+    response.set_cookie "foo", value: "bar", path: "/a", domain: "example.com"
+    response.set_cookie "foo", value: "bar", path: "/a/b", domain: "example.com"
+    response["Set-Cookie"].must_equal [
+      "foo=bar; path=/",
+      "foo=bar; path=/a",
+      "foo=bar; path=/a/b",
+      "foo=bar; domain=example.com.example.com; path=/",
+      "foo=bar; domain=example.com.example.com; path=/a",
+      "foo=bar; domain=example.com.example.com; path=/a/b",
+      "foo=bar; domain=example.com; path=/",
+      "foo=bar; domain=example.com; path=/a",
+      "foo=bar; domain=example.com; path=/a/b",
+    ].join("\n")
+
+    response.delete_cookie "foo", path: "/a", domain: "example.com"
+    response["Set-Cookie"].must_equal [
+      "foo=bar; path=/",
+      "foo=bar; path=/a",
+      "foo=bar; path=/a/b",
+      "foo=bar; domain=example.com.example.com; path=/",
+      "foo=bar; domain=example.com.example.com; path=/a",
+      "foo=bar; domain=example.com.example.com; path=/a/b",
+      "foo=bar; domain=example.com; path=/",
+      "foo=bar; domain=example.com; path=/a/b",
+      "foo=; domain=example.com; path=/a; max-age=0; expires=Thu, 01 Jan 1970 00:00:00 GMT",
+    ].join("\n")
+
+    response.delete_cookie "foo", path: "/a/b", domain: "example.com"
+    response["Set-Cookie"].must_equal [
+      "foo=bar; path=/",
+      "foo=bar; path=/a",
+      "foo=bar; path=/a/b",
+      "foo=bar; domain=example.com.example.com; path=/",
+      "foo=bar; domain=example.com.example.com; path=/a",
+      "foo=bar; domain=example.com.example.com; path=/a/b",
+      "foo=bar; domain=example.com; path=/",
+      "foo=; domain=example.com; path=/a; max-age=0; expires=Thu, 01 Jan 1970 00:00:00 GMT",
+      "foo=; domain=example.com; path=/a/b; max-age=0; expires=Thu, 01 Jan 1970 00:00:00 GMT",
+    ].join("\n")
+  end
+
   it "can do redirects" do
     response = Rack::Response.new
     response.redirect "/foo"
@@ -298,6 +391,33 @@ describe Rack::Response do
     status.must_equal 404
   end
 
+  it "correctly updates Content-Type when writing when not initialized with body" do
+    r = Rack::Response.new
+    r.write('foo')
+    r.write('bar')
+    r.write('baz')
+    _, header, body = r.finish
+    str = "".dup; body.each { |part| str << part }
+    str.must_equal "foobarbaz"
+    header['Content-Length'].must_equal '9'
+  end
+
+  it "correctly updates Content-Type when writing when initialized with body" do
+    obj = Object.new
+    def obj.each
+      yield 'foo'
+      yield 'bar'
+    end
+    ["foobar", ["foo", "bar"], obj].each do
+      r = Rack::Response.new(["foo", "bar"])
+      r.write('baz')
+      _, header, body = r.finish
+      str = "".dup; body.each { |part| str << part }
+      str.must_equal "foobarbaz"
+      header['Content-Length'].must_equal '9'
+    end
+  end
+
   it "doesn't return invalid responses" do
     r = Rack::Response.new(["foo", "bar"], 204)
     _, header, body = r.finish
@@ -400,12 +520,14 @@ describe Rack::Response do
 
   it "provide access to the HTTP headers" do
     res = Rack::Response.new
-    res["Content-Type"] = "text/yaml"
+    res["Content-Type"] = "text/yaml; charset=UTF-8"
 
     res.must_include "Content-Type"
-    res.headers["Content-Type"].must_equal "text/yaml"
-    res["Content-Type"].must_equal "text/yaml"
-    res.content_type.must_equal "text/yaml"
+    res.headers["Content-Type"].must_equal "text/yaml; charset=UTF-8"
+    res["Content-Type"].must_equal "text/yaml; charset=UTF-8"
+    res.content_type.must_equal "text/yaml; charset=UTF-8"
+    res.media_type.must_equal "text/yaml"
+    res.media_type_params.must_equal "charset" => "UTF-8"
     res.content_length.must_be_nil
     res.location.must_be_nil
   end
@@ -495,6 +617,31 @@ describe Rack::Response do
 
     res.finish.flatten.must_be_kind_of(Array)
   end
+
+  it "should specify not to cache content" do
+    response = Rack::Response.new
+
+    response.cache!(1000)
+    response.do_not_cache!
+
+    expect(response['Cache-Control']).must_equal "no-cache, must-revalidate"
+
+    expires_header = Time.parse(response['Expires'])
+    expect(expires_header).must_be :<=, Time.now
+  end
+
+  it "should specify to cache content" do
+    response = Rack::Response.new
+
+    duration = 120
+    expires = Time.now + 100 # At least this far into the future
+    response.cache!(duration)
+
+    expect(response['Cache-Control']).must_equal "public, max-age=120"
+
+    expires_header = Time.parse(response['Expires'])
+    expect(expires_header).must_be :>=, expires
+  end
 end
 
 describe Rack::Response, 'headers' do
